@@ -1,5 +1,5 @@
 ;;  -*-  indent-tabs-mode:nil; coding: utf-8 -*-
-;;  Copyright (C) 2014,2015
+;;  Copyright (C) 2014,2015,2016,2017
 ;;      "Mu Lei" known as "NalaGinrut" <NalaGinrut@gmail.com>
 ;;  Artanis is free software: you can redistribute it and/or modify
 ;;  it under the terms of the GNU General Public License and GNU
@@ -33,8 +33,10 @@
   #:use-module (artanis env)
   #:use-module (artanis mime)
   #:use-module (artanis upload)
+  #:use-module (artanis websocket)
   #:use-module (artanis third-party json)
   #:use-module (artanis third-party csv)
+  #:use-module (artanis server scheduler)
   #:use-module (ice-9 regex) ; FIXME: should use irregex!
   #:use-module (ice-9 match)
   #:use-module (srfi srfi-1)
@@ -64,24 +66,27 @@
             :mime
             :auth
             :session
-            :from-post))
+            :from-post
+            :websocket))
 
 (define (define-handler method rule opts-and-handler)
   (let ((keys (rule->keys rule))
         (path-regexp (compile-rule rule))
         (opts (oah->opts opts-and-handler))
         (handler (oah->handler opts-and-handler)))
+    (when (assoc-ref opts #:websocket)
+      (this-rule-enabled-websocket! rule))
     (hash-set! *handlers-table*
-               (string-append method " " path-regexp)
+               (cons method path-regexp)
                (make-handler-rc handler keys (new-oht opts #:rule rule #:keys keys)))))
 
-(define (get rule . opts-and-handler) (define-handler "GET" rule opts-and-handler))
-(define (post rule . opts-and-handler) (define-handler "POST" rule opts-and-handler))
-(define (put rule . opts-and-handler) (define-handler "PUT" rule opts-and-handler))
-(define (patch rule . opts-and-handler) (define-handler "PATCH" rule opts-and-handler))
+(define (get rule . opts-and-handler) (define-handler 'GET rule opts-and-handler))
+(define (post rule . opts-and-handler) (define-handler 'POST rule opts-and-handler))
+(define (put rule . opts-and-handler) (define-handler 'PUT rule opts-and-handler))
+(define (patch rule . opts-and-handler) (define-handler 'PATCH rule opts-and-handler))
 ;; NOTE: delete method is rarely used, and this name will override the list deletion in Scheme.
 ;;       So my vote is changing its name, feel free to let me know if it's improper.
-(define (page-delete rule . opts-and-handler) (define-handler "DELETE" rule opts-and-handler))
+(define (page-delete rule . opts-and-handler) (define-handler 'DELETE rule opts-and-handler))
 
 ;; NOTE: we banned "\" in the path to avoid SQL-injection!!!
 (define *rule-regexp* (make-regexp ":[^\\/]+"))    
@@ -111,9 +116,7 @@
     (if h (h rc args ...) values)))
 
 (define-syntax-rule (auth-enabled? rc)
-  (begin
-    (format #t "adfa~%")
-    (hash-ref (get-oht rc) #:auth)))
+  (hash-ref (get-oht rc) #:auth))
 ;; --------------------------------------------------------------
 ;; oht handlers
 
@@ -141,7 +144,8 @@
         ('all (DB-get-all-rows r))
         ('top (DB-get-top-row r))
         ((? positive? n) (DB-get-n-rows r n))
-        (else (throw 'artanis-err 500 "raw-sql: Invalid mode!" mode))))))
+        (else (throw 'artanis-err 500 raw-sql-maker
+                     "Invalid mode!" mode))))))
 
 ;; for #:cache
 (define (cache-maker pattern rule keys)
@@ -164,7 +168,8 @@
        (try-to-cache-static-file rc file 'private maxage)))
     ((or (? boolean? opts) (? list? opts))
      (lambda (rc body) (try-to-cache-body rc body opts)))
-    (else (throw 'artanis-err "cache-maker: invalid pattern!" pattern))))
+    (else (throw 'artanis-err 500 cache-maker
+                 "Invalid pattern!" pattern))))
 
 ;; for #:cookies
 ;; NOTE: Support cookies customization, which let users define special cookies.
@@ -178,7 +183,7 @@
   (define ckl '())
   (define (cget ck)
     (or (assoc-ref ckl ck)
-        (throw 'artansi-err 500 "Undefined cookie name" ck)))
+        (throw 'artansi-err 500 cget "Undefined cookie name" ck)))
   (define (cset! ck k v)
     (and=> (cget ck) (cut %cookie-set! <> k v)))
   (define (cref ck k)
@@ -205,7 +210,8 @@
      (set! %cookie-ref getter)
      (set! %cookie-modify modifier)
      (init-cookies names))
-    (else (throw 'artansi-err 500 "cookies-maker: Invalid mode while init!" mode)))
+    (else (throw 'artansi-err 500 cookies-maker
+                 "Invalid mode while init!" mode)))
   (lambda (rc op)
     (case op
       ((set) cset!)
@@ -213,20 +219,21 @@
       ((setattr) setattr)
       ((update) update)
       ((remove) remove)
-      (else (throw 'artanis-err 500 "cookies-maker: Invalid operation!" op)))))
+      (else (throw 'artanis-err 500 cookies-maker
+                   "Invalid operation!" op)))))
 
 ;; for #:mime
 (define (mime-maker type rule keys)
-  (define mime (mime-guess type))
-  (lambda (rc . args)
-    (define headers `((content-type . (,(mime-guess type)))))
-    (define-syntax-rule (-> func) (values (apply func args) #:pre-headers headers))
-    (case type
-      ((json jsonp) (-> ->json-string))
-      ((xml) (-> sxml->xml-string))
-      ((csv) (-> sxml->csv-string))
-      ((sxml) (values (object->string (car args)) #:pre-headers headers))
-      (else (throw 'artanis-err 500 "mime-maker: Invalid type!" type)))))
+  (define headers `((content-type . (,(mime-guess type)))))
+  (define-syntax-rule (-> func the-args) (values (apply func the-args) #:pre-headers headers))
+  (define gen-mime
+   (case type
+     ((json jsonp) (lambda (args) (-> ->json-string args)))
+     ((xml) (lambda (args) (-> sxml->xml-string args)))
+     ((csv) (lambda (args) (-> sxml->csv-string args)))
+     ((sxml) (lambda (args) (values (object->string (car args)) #:pre-headers headers)))
+     (else (throw 'artanis-err 500 mime-maker "Invalid type!" type))))
+  (lambda (rc . args) (gen-mime args)))
 
 ;; for #:session
 (define (session-maker mode rule keys)
@@ -237,15 +244,16 @@
         (let ((cookie (new-cookie #:npv `((,idname . ,sid)))))
           (and cookie (rc-set-cookie! rc (list cookie)))
           sid))))
-  (define* (spawn-handler rc #:key (keep? #f))
+  (define spawn-handler
     (match mode
-      ((or #t 'spawn) (%spawn rc))
-      (`(spawn ,sid) (%spawn rc #:idname sid))
-      (`(spawn ,sid ,proc) (%spawn rc #:idname sid #:proc proc))
-      (else (throw 'artanis-err 500 "session-maker: Invalid config mode" mode))))
+      ((or #t 'spawn) %spawn)
+      (`(spawn ,sid) (lambda (rc) (%spawn rc #:idname sid)))
+      (`(spawn ,sid ,proc) (lambda (rc) (%spawn rc #:idname sid #:proc proc)))
+      (else (throw 'artanis-err 500 session-maker "Invalid config mode: ~a" mode))))
   (define (check-it rc idname)
-    (and=> (session-restore (cookie-ref (rc-cookie rc) idname))
-           (lambda (s) (session-from-correct-client? s rc))))
+    (let ((sid (any (lambda (c) (cookie-ref c idname)) (rc-cookie rc))))
+      (and=> (session-restore (or sid ""))
+             (lambda (s) (session-from-correct-client? s rc)))))
   (lambda (rc cmd)
     (match cmd
       ('check (check-it rc "sid"))
@@ -255,17 +263,14 @@
       (`(check-and-spawn ,sid)
        (or (check-it rc sid) (spawn-handler rc)))
       ('spawn (spawn-handler rc))
-      (else (throw 'artanis-err 500 "session-maker: Invalid call cmd" cmd)))))
+      (else (throw 'artanis-err 500 session-maker "Invalid call cmd: ~a" cmd)))))
 
+;; for #:from-post
 (define (from-post-maker mode rule keys)
-  (define post-data #f)
   (define* (post->qstr-table rc #:optional (safe? #f))
-    (when (not post-data)
-      (set! post-data
-            (if (rc-body rc)
-                (generate-kv-from-post-qstr (rc-body rc) #:no-evil? safe?)
-                '())))
-    post-data)
+    (if (rc-body rc)
+        (generate-kv-from-post-qstr (rc-body rc) #:no-evil? safe?)
+        '()))
   (define (default-success-ret size-list filename-list)
     (call-with-output-string
      (lambda (port)
@@ -289,8 +294,8 @@
       (`(success ,mfds ,slist ,flist) (success-ret mfds slist flist))
       ('none (no-file-ret))
       (else
-       (throw 'artanis-err 500
-              "store-the-bv: Impossible status! please report bug!"))))
+       (throw 'artanis-err 500 store-the-bv
+              "Impossible status! please report bug!"))))
   (define (post-ref plst key) (and=> (assoc-ref plst key) car))
   (define (post-handler rc)
     (match mode
@@ -298,14 +303,76 @@
       ('qstr-safe (post->qstr-table rc 'safe))
       ((or 'bv 'bytevector) (rc-body rc))
       ;; upload operation, indeed
-      (('store rest ...) (apply store-the-bv rc rest))
-      (else (throw 'artanis-err 500 "post-handler: Invalid mode!" mode))))
+      (('store rest ...)
+       (catch 'system-error
+         (lambda ()
+           (apply store-the-bv rc rest))
+         (lambda e
+           (let ((errno (system-error-errno e)))
+             (cond
+              ((= errno ENOMEM)
+               ;; NOTE: Out of memory, call (gc) and throw 507
+               (format (artanis-current-output) "No memory! Run GC now!~%")
+               (gc)
+               (throw 'artanis-err 507 post-handler
+                      "Server is out of RAMs, please extend more RAMs!~%"))
+              ((= errno EIO)
+               ;; NOTE: The storage device was disconnected, just throw 507
+               (throw 'artanis-err 507 post-handler
+                      "Server is not available, maybe storage media was disconnected?~%"))
+              ((= errno ENOSPC)
+               ;; NOTE: no space for uploading, just throw 507
+               (throw 'artanis-err 507 post-handler
+                      "Server has insufficient storage space!~%"))
+              (else
+               ;; nothing noticed, re-throw it to next level.
+               (apply throw e)))))))
+      (else (throw 'artanis-err 500 post-handler "Invalid mode!" mode))))
+  (define (get-values-from-post pl . keys)
+    (apply
+     values
+     (map
+      (lambda (k) (post-ref pl k))
+      keys)))
   (lambda (rc . cmd)
     (match cmd
       (`(get ,key) (post-ref (post-handler rc) key))
-      ('(get) (rc-body rc))
+      ('(get) (post-handler rc))
+      (('get-vals keys ...) (apply get-values-from-post (post-handler rc) keys))
       ('(store) (post-handler rc))
-      (else (throw 'artanis-err 500 "from-post-maker: Invalid cmd!" cmd)))))
+      (else (throw 'artanis-err 500 from-post-maker "Invalid cmd!" cmd)))))
+
+;; for #:websocket
+;; 
+(define (websocket-maker type rule keys)
+  (define (call-with-websocket-channel args)
+    (match args
+      (('proto (? symbol? proto))
+       ;; TODO: call protocol initilizer, and establish websocket for it.
+       #t)
+      (('redirect (? string? ip/usk))
+       ;; NOTE: We use IP rather than hostname, since it's usually redirected to
+       ;;       a LAN address. Using hostname may cause DNS issues.
+       ;; NOTE: ip/usk means ip or unix-socket, the pattern should be this:
+       ;;       ^ip://(?:[0-9]{1,3}\\.){3}[0-9]{1,3}(:[0-9]{1,5})?$
+       ;;       ^unix://[a-zA-Z-_0-9]+\\.socket$
+       ;; NOTE: Equivalent to transparent proxy.
+       ;; TODO: call protocol initilizer, and establish websocket
+       ;;       to redirect it.
+       #t)
+      (('proxy (? symbol? proto))
+       ;; NOTE: Setup a proxy with certain protocol handler.
+       ;;       Different from the regular proxy design, the proxy in Artanis doesn't
+       ;;       need a listen port, since it's always 80/443. The client should
+       ;;       support websocket, and visit the related URL for establishing
+       ;;       a websocket channel. Then the rest is the same with regular proxy.
+       #t)
+      (else (throw 'artanis-err 500 websocket-maker "Invalid type!" type))))
+  (this-rule-enabled-websocket! rule)
+  (lambda (rc . args)
+    (error "Artanis hasn't been ready for websocket yet!")
+    ;; TODO: parsing command and apply call-with-websocket-channel
+    #t))
 
 ;; ---------------------------------------------------------------------------------
   
@@ -316,6 +383,8 @@
 ;; CONVENTION:
 ;; 1. When a var is arounded with ${..}, and there's ":" as its prefix,
 ;;    say, ${:xxx}, it means "to get xxx from rule key".
+;; 2. Arounded with ${..}, and there's "@" as its prefix, say, ${@xxx},
+;;    it means "to get xxx from body" (POST way).
 (define *options-meta-handler-table*
   `(;; a short way for sql-mapping from the bindings in rule
     ;; e.g #:sql-mapping "select * from table where id=${:id}"
@@ -427,7 +496,31 @@
     (#:session . ,session-maker)
 
     ;; Get data from POST
-    (#:from-post . ,from-post-maker)))
+    (#:from-post . ,from-post-maker)
+
+    ;; Establish websocket channel
+    ;; Each time developers set :websocket and specifed a protocol, say, `myproto',
+    ;; Artanis will check if the file app/protocols/myproto.scm exists, then load
+    ;; the protocol initialize function in `myproto' which will also add `myproto'
+    ;; into *proto-table*. Then ragnarok will take charge of it to call the correct
+    ;; handlers.
+    ;; There're 3 cases:
+    ;; 1. Regular protocol handling over websocket
+    ;;    The initializer will create an instance for the protocol.
+    ;; 2. Redirect to another address (IP or UNIX socket)
+    ;;    The initializer will create a binding pair in redirect table:
+    ;;    src port and des port.
+    ;;    In this case, there's no protocol handling in Artanis. The protocol handler
+    ;;    is in remote (could be a host machine, or a process in current OS).
+    ;; 3. Proxy mode:
+    ;;    Different from reguar proxy, the proxy in Artanis runs over websocket, and
+    ;;    there's no extra listenning port, only 80/443.
+    ;;    In this case, there're 2 situations:
+    ;;    a. transparent proxy
+    ;;       No protocol handling at all, just redirecting the data without cooking.
+    ;;    b. cooked proxy
+    ;;       There's a bound protocol instance for it.
+    (#:websocket . ,websocket-maker)))
 
 (define-macro (meta-handler-register what)
   `(define-syntax-rule (,(symbol-append ': what) rc args ...)
@@ -445,6 +538,7 @@
 (meta-handler-register auth)
 (meta-handler-register session)
 (meta-handler-register from-post)
+(meta-handler-register websocket)
 
 (define-syntax-rule (:cookies-set! rc ck k v)
   ((:cookies rc 'set) ck k v))
