@@ -1,5 +1,5 @@
 ;;  -*-  indent-tabs-mode:nil; coding: utf-8 -*-
-;;  Copyright (C) 2013,2014,2015
+;;  Copyright (C) 2013,2014,2015,2016,2017
 ;;      "Mu Lei" known as "NalaGinrut" <NalaGinrut@gmail.com>
 ;;  Artanis is free software: you can redistribute it and/or modify
 ;;  it under the terms of the GNU General Public License and GNU
@@ -37,15 +37,22 @@
   #:use-module (ice-9 local-eval)
   #:use-module (ice-9 receive)
   #:use-module (ice-9 q)
+  #:use-module (ice-9 control)
   #:use-module (web http)
   #:use-module (web request)
+  #:use-module (web uri)
   #:use-module ((rnrs)
                 #:select (get-bytevector-all utf8->string put-bytevector
                           bytevector-u8-ref string->utf8 bytevector-length
-                          make-bytevector bytevector-s32-native-ref))
-  #:export (regexp-split hash-keys cat bv-cat get-global-time
-            get-local-time string->md5 unsafe-random string-substitute
-            get-file-ext get-global-date get-local-date uri-decode
+                          make-bytevector bytevector-s32-native-ref bytevector?
+                          define-record-type record-rtd record-accessor
+                          get-string-all))
+  #:export (regexp-split hash-keys cat bv-cat get-global-time sanitize-response
+            build-response write-response get-local-time string->md5 unsafe-random
+            uri-decode response-version response-code response-connection
+            response-port write-response-body read-request request-uri request-method
+            request-content-length request-port read-request-body response-content-length
+            get-file-ext get-global-date get-local-date string-substitute
             nfx static-filename remote-info seconds-now local-time-stamp
             parse-date write-date make-expires export-all-from-module!
             alist->hashtable expires->time-utc local-eval-string
@@ -63,15 +70,21 @@
             sxml->xml-string run-after-request! run-before-response!
             make-pipeline HTML-entities-replace eliminate-evil-HTML-entities
             generate-kv-from-post-qstr handle-proper-owner
-            generate-data-url verify-ENTRY
+            generate-data-url verify-ENTRY exclude-dbd
             draw-expander remove-ext scan-app-components cache-this-route!
             dump-route-from-cache generate-modify-time delete-directory
-            handle-existing-file check-drawing-method
+            handle-existing-file check-drawing-method DEBUG
             subbv->string subbv=? bv-read-line bv-read-delimited put-bv
             bv-u8-index bv-u8-index-right build-bv-lookup-table filesize
             plist-remove gen-migrate-module-name try-to-load-migrate-cache
-            flush-to-migration-cache gen-local-conf-file with-dbd errno
-            c/struct-sizeof)
+            flush-to-migration-cache gen-local-conf-file with-dbd
+            call-with-sigint define-box-type make-box-type unbox-type
+            ::define did-not-specify-parameter WARN-TEXT ERROR-TEXT REASON-TEXT
+            NOTIFY-TEXT STATUS-TEXT get-trigger get-family get-addr request-path
+            keep-alive? procedure-name->string proper-toplevel gen-content-length
+            make-file-sender file-sender? file-sender-size file-sender-thunk
+            get-string-all-with-detected-charset make-unstop-exception-handler
+            artanis-log exception-from-client exception-from-server render-sys-page)
   #:re-export (the-environment))
 
 ;; There's a famous rumor that 'urandom' is safer, so we pick it.
@@ -87,6 +100,20 @@
 (define uri-decode (@ (web uri) uri-decode))
 (define parse-date (@@ (web http) parse-date))
 (define write-date (@@ (web http) write-date))
+(define build-response (@ (web response) build-response))
+(define write-response (@ (web response) write-response))
+(define response-version (@ (web response) response-version))
+(define response-code (@ (web response) response-code))
+(define response-connection (@ (web response) response-connection))
+(define response-port (@ (web response) response-port))
+(define write-response-body (@ (web response) write-response-body))
+(define response-content-length (@ (web response) response-content-length))
+(define read-request (@ (web request) read-request))
+(define read-request-body (@ (web request) read-request-body))
+(define request-uri (@ (web request) request-uri))
+(define request-method (@ (web request) request-method))
+(define request-content-length (@ (web request) request-content-length))
+(define request-port (@ (web request) request-port))
 
 (define-syntax-rule (local-eval-string str e)
   (local-eval 
@@ -101,10 +128,10 @@
     ht))
 
 (eval-when (eval load compile)
-           (define (export-all-from-module! module-name)
-             (let ((mod (resolve-module module-name)))
-               (module-for-each (lambda (s m) 
-                                  (module-add! (current-module) s m)) mod))))
+  (define (export-all-from-module! module-name)
+    (let ((mod (resolve-module module-name)))
+      (module-for-each (lambda (s m)
+                         (module-add! (current-module) s m)) mod))))
 
 (define (time-expired? expires)
   (if expires
@@ -455,8 +482,7 @@
             item
             (or (and=> (get-the-val keyword-args (car item)) ->string)
                 (cdr item) ; default value
-                (throw 'artanis-err 500
-                       "(utils)item->string: Missing keyword" (car item)))))
+                "")))
       (string-concatenate (map item->string items)))))
 
 ;; the normal mode, no double quotes for vals
@@ -910,7 +936,7 @@
           (lp (cddr next) (list (car next) (cadr next)) ret)))
      (else (lp (cdr next) kk (cons (car next) ret))))))
 
-(define *name-re* (string->sre "([^.]+)\\.scm"))
+(define *name-re* (string->irregex "([^.]+)\\.scm"))
 (define (gen-migrate-module-name f)
   (cond
    ((irregex-search *name-re* (basename f))
@@ -952,39 +978,337 @@
                      dbd0 dbd1)
              'body ...)))))
 
-(define %libc-errno-pointer
-  ;; Glibc's 'errno' pointer.
-  (let ((errno-loc (dynamic-func "__errno_location" (dynamic-link))))
-    (and errno-loc
-         (let ((proc (pointer->procedure '* errno-loc '())))
-           (proc)))))
+(define-syntax-rule (exclude-dbd dbds body ...)
+  (let ((dbd (get-conf '(db dbd))))
+    (cond
+     ((memq dbd dbds)
+      (throw 'artanis-err 500
+             (format #f "This isn't supported by `~a', please check it out again!"
+                     dbds)
+             'body ...))
+     (else body ...))))
 
-(define errno
-  (if %libc-errno-pointer
-      (let ((bv (pointer->bytevector %libc-errno-pointer (sizeof int))))
-        (lambda ()
-          "Return the current errno."
-          ;; XXX: We assume that nothing changes 'errno' while we're doing all this.
-          ;; In particular, that means that no async must be running here.
+(define-syntax-rule (DEBUG fmt args ...)
+  (when (get-conf 'debug-mode)
+        (format (artanis-current-output) fmt args ...)))
 
-          ;; Use one of the fixed-size native-ref procedures because they are
-          ;; optimized down to a single VM instruction, which reduces the risk
-          ;; that we fiddle with 'errno' (needed on Guile 2.0.5, libc 2.11.)
-          (let-syntax ((ref (lambda (s)
-                              (syntax-case s ()
-                                ((_ bv)
-                                 (case (sizeof int)
-                                   ((4)
-                                    #'(bytevector-s32-native-ref bv 0))
-                                   ((8)
-                                    #'(bytevector-s64-native-ref bv 0))
-                                   (else
-                                    (error "unsupported 'int' size"
-                                           (sizeof int)))))))))
-            (ref bv))))
-      (lambda () 0)))
+(define call-with-sigint
+  (if (not (provided? 'posix))
+      (lambda (thunk handler-thunk) (thunk))
+      (lambda (thunk handler-thunk)
+        (let ((handler #f))
+          (catch 'interrupt
+            (lambda ()
+              (dynamic-wind
+                (lambda ()
+                  (set! handler
+                        (sigaction SIGINT (lambda (sig) (throw 'interrupt)))))
+                thunk
+                (lambda ()
+                  (if handler
+                      ;; restore Scheme handler, SIG_IGN or SIG_DFL.
+                      (sigaction SIGINT (car handler) (cdr handler))
+                      ;; restore original C handler.
+                      (sigaction SIGINT #f)))))
+            (lambda (k . _) (handler-thunk)))))))
 
-(define (c/struct-sizeof meta)
-  (apply +
-         (map (lambda (m) (if (list? m) (c/struct-sizeof m) (sizeof m)))
-              meta)))
+(define-syntax-rule (define-box-type name)
+  (define-record-type name (fields treasure)))
+
+(define-macro (make-box-type bt v)
+  (list (symbol-append 'make- bt) v))
+
+(define-syntax-rule (box-type-treasure t)
+  (record-accessor (record-rtd t) 0))
+
+(define-syntax-rule (unbox-type t)
+  (let ((treasure-getter (box-type-treasure t)))
+    (treasure-getter t)))
+
+(define (socket-port? sp)
+  (and (port? sp) (eq? (port-filename sp) 'socket)))
+
+(define (detect-type-name o)
+  (define r6rs-record? (@ (rnrs) record?))
+  (define r6rs-record-type-name (@ (rnrs) record-type-name))
+  (define guile-specific-record? (@ (guile) record?))
+  (define (guile-specific-record-name o)
+    ((@ (guile) record-type-name)
+     ((@ (guile) record-type-descriptor) o)))
+  (cond
+   ;; NOTE: record? is for R6RS record-type, but record-type? is not
+   ((r6rs-record? o) (r6rs-record-type-name (record-rtd o)))
+   ((guile-specific-record? o) (guile-specific-record-name o))
+   ((symbol? o) 'symbol)
+   ((string? o) 'string)
+   ((integer? o) 'int)
+   ((number? o) 'number)
+   ((thunk? o) 'thunk)
+   ((procedure? o) 'proc)
+   ((vector? o) 'vector)
+   ((pair? o) 'pair)
+   ((list? o) 'list)
+   ((bytevector? o) 'bv)
+   ((socket-port? o) 'socket)
+   ((boolean? o) 'boolean)
+   (else 'ANY)))
+
+(define (check-args-types op args)
+  (match (procedure-property op 'type-anno)
+    (((targs ...) '-> (func-types ...))
+     (for-each
+      (lambda (v e)
+        (or (eq? e 'ANY)
+            (eq? (detect-type-name v) e)
+            (begin
+              (DEBUG "(~{~a~^ ~}) =? (~{~a~^ ~})~%" targs args)
+              (throw 'artanis-err 500
+                     (format #f "~a: Argument ~a is a `~a' type, but I expect type `~a'"
+                             op v (detect-type-name v) e)))))
+      args targs))
+    (else (throw 'artanis-err 500 "Invalid type annotation"
+                 (procedure-property op 'type-anno)))))
+
+(define (check-function-types op fret)
+  (match (procedure-property op 'type-anno)
+    (((targs ...) '-> (func-types ...))
+     (for-each
+      (lambda (v e)
+        (or (eq? e 'ANY)
+            (eq? (detect-type-name v) e)
+            (throw 'artanis-err 500
+                   (format #f "`Return value ~a(~a) is expected to be type `~a'"
+                           v (detect-type-name v) e))))
+      fret func-types))
+    (else (throw 'artanis-err 500 "Invalid type annotation"
+                 (procedure-property op 'type-anno)))))
+
+(define (detect-and-set-type-anno! op ftypes atypes)
+  (let ((type `(,atypes -> ,ftypes)))
+    (set-procedure-property! op 'type-anno type)
+    type))
+
+;; NOTE: This macro can detect multi return values.
+;; TODO:
+;; 1. support multi-types, say, string/bv, maybe not easy to do it faster?
+(define-syntax ::define
+  (syntax-rules (-> :anno:)
+    ((_ (op args ...) (:anno: (targs ...) -> func-types ...) body ...)
+     (begin
+       (define (op args ...)
+         (when (get-conf 'debug-mode) (check-args-types op (list args ...)))
+         (call-with-values
+             (lambda () body ...)
+           (lambda ret
+             (when (get-conf 'debug-mode)
+                   (eq? (detect-type-name ret) (check-function-types op ret)))
+             (apply values ret))))
+       (detect-and-set-type-anno! op '(func-types ...) '(targs ...))))))
+
+(define-syntax-rule (did-not-specify-parameter what)
+  (format #f "`current-~a' isn't specified, it's likely a bug!" what))
+
+;; Text-coloring helper functions, borrowed from guile-colorized
+(define *color-list*
+  `((CLEAR       .   "0")
+    (RESET       .   "0")
+    (BOLD        .   "1")
+    (DARK        .   "2")
+    (UNDERLINE   .   "4")
+    (UNDERSCORE  .   "4")
+    (BLINK       .   "5")
+    (REVERSE     .   "6")
+    (CONCEALED   .   "8")
+    (BLACK       .  "30")
+    (RED         .  "31")
+    (GREEN       .  "32")
+    (YELLOW      .  "33")
+    (BLUE        .  "34")
+    (MAGENTA     .  "35")
+    (CYAN        .  "36")
+    (WHITE       .  "37")
+    (ON-BLACK    .  "40")
+    (ON-RED      .  "41")
+    (ON-GREEN    .  "42")
+    (ON-YELLOW   .  "43")
+    (ON-BLUE     .  "44")
+    (ON-MAGENTA  .  "45")
+    (ON-CYAN     .  "46")
+    (ON-WHITE    .  "47")))
+
+(define (get-color color)
+  (assq-ref *color-list* color))
+
+(define (generate-color colors)
+  (let ((color-list
+         (filter-map get-color colors)))
+    (if (null? color-list)
+        ""
+        (string-append "\x1b[" (string-join color-list ";" 'infix) "m"))))
+
+(define* (colorize-string-helper color str control #:optional (rl-ignore #f))
+  (if rl-ignore
+      (string-append "\x01" (generate-color color) "\x02" str "\x01" (generate-color control) "\x02")
+      (string-append (generate-color color) str (generate-color control))))
+
+(define* (colorize-string str color)
+  "Example: (colorize-string \"hello\" '(BLUE BOLD))" 
+  (colorize-string-helper color str '(RESET) (using-readline?)))
+
+(define-syntax-rule (WARN-TEXT str) (colorize-string str '(YELLOW)))
+(define-syntax-rule (ERROR-TEXT str) (colorize-string str '(RED)))
+(define-syntax-rule (REASON-TEXT str) (colorize-string str '(CYAN)))
+(define-syntax-rule (NOTIFY-TEXT str) (colorize-string str '(WHITE)))
+(define-syntax-rule (STATUS-TEXT num) (colorize-string (object->string num)'(WHITE)))
+
+(define (get-trigger)
+  (case (get-conf '(server trigger))
+    ((edge) (@ (artanis server epoll) EPOLLET))
+    ((level) 0)
+    (else (throw 'artanis-err 500 get-trigger "Invalid (server trigger)!"
+                 (get-conf '(server trigger))))))
+
+(define (get-family)
+  (case (get-conf '(host family))
+    ((ipv4) AF_INET)
+    ((ipv6) AF_INET6)
+    (else (throw 'artanis-err 500 get-family "Invalid (host family)!"
+                 (get-conf '(host family))))))
+        
+(define (get-addr)
+  (let ((host (get-conf '(host addr)))
+        (family (get-family)))
+    (inet-pton family host)))
+
+(define (request-path req)
+  (uri-path (request-uri req)))
+
+(define (keep-alive? response)
+  (let ((v (response-version response)))
+    (and (or (< (response-code response) 400)
+             (= (response-code response) 404))
+         (case (car v)
+           ((1)
+            (case (cdr v)
+              ;; NOTE: HTTP/1.1 treat all connection keep-alive
+              ;;       unless it requires `close' explicityly
+              ((1) (not (memq 'close (response-connection response))))
+              ;; HTTP/1.0 needs explicit keep-alive notice
+              ((0) (memq 'keep-alive (response-connection response)))))
+           (else #f)))))
+
+(define (procedure-name->string proc)
+  (symbol->string (procedure-name proc)))
+
+(define-syntax-rule (proper-toplevel)
+  (or (current-toplevel) ""))
+
+(define-record-type file-sender
+  (fields size thunk))
+
+(define (gen-content-length body)
+  (let ((get-length (lambda ()
+                      (cond
+                       ((bytevector? body) (bytevector-length body))
+                       ((file-sender? body) (file-sender-size body))
+                       ((string? body)
+                        (throw 'artanis-err 500 gen-content-length
+                               "BUG: body should have been converted to bytevector! ~a"
+                               body))
+                       (else (throw 'artanis-err 500 gen-content-length
+                                    "Invalid body ~a" body))))))
+    `(content-length . ,(if body (get-length) 0))))
+
+(define (get-string-all-with-detected-charset filename)
+  (call-with-input-file filename
+    (lambda (port)
+      (set-port-encoding! port (get-conf '(server charset)))
+      (get-string-all port))))
+
+(define (syspage-show file)
+  (let ((local-syspage (format #f "~a/sys/pages/~a" (current-toplevel) file)))
+    (cond
+     ((file-exists? local-syspage)
+      (bv-cat local-syspage #f))
+     (else
+      (bv-cat (string-append (get-conf '(server syspage path)) "/" file) #f)))))
+
+;; ENHANCE: use colored output
+(define* (artanis-log blame-who? status mime #:key (port (current-error-port))
+                      (request #f))
+  (case blame-who?
+    ((client)
+     (when (not request)
+       (error "artanis-log: Fatal bug! Request shouldn't be #f here!~%"))
+     (let* ((uri (request-uri request))
+            (path (uri-path uri))
+            (qstr (uri-query uri))
+            (method (request-method request)))
+       (format port "[Remote] ~a @ ~a~%" (remote-info request) (local-time-stamp))
+       (format port "[Request] method: ~a, path: ~a, qeury: ~a~%" method path qstr)
+       (format port "[Response] status: ~a, MIME: ~a~%~%" status mime)))
+    ((server)
+     (format port "[Server] ~a @ ~a~%" (get-conf '(host addr)) (local-time-stamp))
+     (format port "[Response] status: ~a, MIME: ~a~%~%" status mime))
+    (else (error "artanis-log: Fatal BUG here!"))))
+
+(define (render-sys-page blame-who? status request)
+  (define-syntax-rule (status->page s)
+    (format #f "~a.html" s))
+  (artanis-log blame-who? status 'text/html #:request request)
+  (let* ((charset (get-conf '(server charset)))
+         (mtime (generate-modify-time (current-time)))
+         (response
+          (build-response #:code status
+                          #:headers `((server . ,(get-conf '(server info)))
+                                      (last-modified . ,mtime)
+                                      (content-type . (text/html (charset . ,charset))))))
+         (body (syspage-show (status->page status))))
+    (if (eq? (get-conf '(server engine)) 'guile)
+        (values response body)
+        (values response body 'exception))))
+
+(define (format-status-page/client status request)
+  (format (current-error-port) "[EXCEPTION] ~a is abnormal request, status: ~a, "
+          (uri-path (request-uri request)) status)
+  (display "rendering a sys page for it...\n" (current-error-port)) 
+  (render-sys-page 'client status request))
+
+(define (format-status-page/server status)
+  (format (current-error-port) "[SERVER ERROR] Internal error from server-side, ")
+  (format (current-error-port) "rendering a ~a page for client ...~%" status)
+  (render-sys-page 'server status #f))
+
+(define (exception-from-client request)
+  (lambda (status)
+    (format-status-page/client status request)))
+
+(define (exception-from-server)
+  (lambda (status)
+    (format-status-page/server status)))
+
+(define (make-unstop-exception-handler syspage-generator)
+  (lambda (k . e)
+    (define port (current-error-port))
+    (format port (ERROR-TEXT "GNU Artanis encountered exception!~%"))
+    (match e
+      (((? procedure? subr) (? string? msg) . args)
+       (format port "<~a>~%" (WARN-TEXT (current-filename)))
+       (when subr (format port "In procedure ~a :~%"
+                          (WARN-TEXT (procedure-name->string subr))))
+       (apply format port (REASON-TEXT msg) args)
+       (newline port))
+      (((? integer? status) (? procedure? subr) (? string? msg) . args)
+       (format port "HTTP ~a~%" (STATUS-TEXT status))
+       (format port "<~a>~%" (WARN-TEXT (current-filename)))
+       (when subr (format port "In procedure ~a :~%"
+                          (WARN-TEXT (procedure-name->string subr))))
+       (apply format port (REASON-TEXT msg) args)
+       (newline port)
+       (syspage-generator status))
+      (else
+       (format port "~a - ~a~%"
+               (WARN-TEXT
+                "BUG: invalid exception format, but we throw it anyway!")
+               e)
+       (apply throw k e)))))
